#include "fncmalloc.h"
#include <dlfcn.h>

/* used to suspend memory tracing (during my own treatments) */
int hx_trace_suspend=0;
int hx_trace_memory=0;
int hx_trace_stack=0;  /* stack-size to be dropped in memory file */
FILE *hx_trace_file=NULL;  /* file for memory trace */
char hx_trace_buf[BUFSIZ];


/* hooks */
typedef void * (mallfunc)(size_t);
typedef void * (callfunc)(size_t, size_t);
typedef void * (rallfunc)(void *, size_t);
typedef void * (mlinfunc)(size_t, size_t);
typedef void (freefunc)(void *);

/* Global variables used to hold actual function addresses.  */
static mallfunc *hx_malloc;
static callfunc *hx_calloc;
static mlinfunc *hx_memalign;
static rallfunc *hx_realloc;
static mallfunc *hx_valloc;
static freefunc *hx_free;

static int hx_malloc_inited;

/* initialisation of malloc's hooks */
void hx_init_malloc()
{
  hx_malloc = dlsym( RTLD_NEXT, "malloc" );
  hx_calloc = dlsym( RTLD_NEXT, "calloc" );
  hx_valloc = dlsym( RTLD_NEXT, "valloc" );
  hx_realloc = dlsym( RTLD_NEXT, "realloc" );
  hx_memalign = dlsym( RTLD_NEXT, "memalign" );
  hx_free = dlsym( RTLD_NEXT, "free" );
  hx_malloc_inited = 1;
}


/* display in the output file the current stack (last elements) */
void hx_draw_stack_memory()
{
  int i, j;
#ifdef HAVE_LIBPTHREAD
  pthread_t pt;
#endif

  /* no stack */
  if (hx_trace_stack == 0)
    return;

#ifdef HAVE_LIBPTHREAD
  if (hx_pstack)
    pt = pthread_self();
#endif

  for(i=hx_pos_in_stack-1,j=0; i>0 && j<hx_trace_stack; i--)
    {
#ifdef HAVE_LIBPTHREAD
    if (hx_pstack && pt != hx_pstack[i])
      continue;
#endif
    if (i == 0)
      {/* element #0 can be an invalid address */
      fprintf(hx_trace_file, "0x1");
      } /*                     ^ flag for invalid address */
    fprintf(hx_trace_file, "%#x ", hx_mstack[i]);
    j++;
    }
  putc('\n', hx_trace_file);
}

/* my own malloc/realloc/free */
void * malloc(size_t size)
{
  void * result, *caller=__builtin_return_address(0);

  HX_LOCK;
  if (!hx_malloc_inited) hx_init_malloc();

  if (hx_malloc) result = hx_malloc(size);
  else result = NULL;

  if (!hx_trace_suspend && hx_trace_file)
    {
    fprintf(hx_trace_file, "+ %#x %#x %#x\n", caller, result, (unsigned int)size);
    hx_draw_stack_memory();
    }

  HX_UNLOCK;

  /* return the pointer */
  return(result);
}

void * calloc(size_t nelem, size_t size)
{
  void * result, *caller=__builtin_return_address(0);

  HX_LOCK;
  if (!hx_malloc_inited) hx_init_malloc();

  if (hx_calloc) result = hx_calloc(nelem, size);
  else result = NULL;

  if (!hx_trace_suspend && hx_trace_file)
    {
    fprintf(hx_trace_file, "+ %#x %#x %#x\n", caller, result, (unsigned int)size*nelem);
    hx_draw_stack_memory();
    }

  HX_UNLOCK;

  /* return the pointer */
  return(result);
}

void * valloc(size_t size)
{
  void * result, *caller=__builtin_return_address(0);

  HX_LOCK;
  if (!hx_malloc_inited) hx_init_malloc();

  if (hx_valloc) result = hx_valloc(size);
  else result = NULL;

  if (!hx_trace_suspend && hx_trace_file)
    {
    fprintf(hx_trace_file, "+ %#x %#x %#x\n", caller, result, (unsigned int)size);
    hx_draw_stack_memory();
    }

  HX_UNLOCK;

  /* return the pointer */
  return(result);
}

void * realloc(void * ptr, size_t size)
{
  void * result, *caller=__builtin_return_address(0);

  HX_LOCK;
  if (!hx_malloc_inited) hx_init_malloc();

  if (hx_realloc) result = hx_realloc(ptr, size);
  else result = NULL;

  /* `printf' might call `malloc', so protect it too. */
  if (!hx_trace_suspend && hx_trace_file)
    {
    fprintf(hx_trace_file, "> %#x %#x %#x %#x\n", caller, ptr, result, size);
    hx_draw_stack_memory();
    }

  HX_UNLOCK;

  /* return the pointer */
  return(result);
}

void * memalign(size_t align, size_t size)
{
  void * result, *caller=__builtin_return_address(0);

  HX_LOCK;
  if (!hx_malloc_inited) hx_init_malloc();

  if (hx_memalign) result = hx_memalign(align, size);
  else result = NULL;

  /* `printf' might call `malloc', so protect it too. */
  if (!hx_trace_suspend && hx_trace_file)
    {
    fprintf(hx_trace_file, "+ %#x %#x %#x\n", caller, result, (unsigned int)size);
    hx_draw_stack_memory();
    }

  HX_UNLOCK;

  /* return the pointer */
  return(result);
}

void free(void * ptr)
{
  void *caller=__builtin_return_address(0);

  HX_LOCK;
  if (!hx_malloc_inited) hx_init_malloc();

  if (hx_free) hx_free(ptr);

  /* `printf' might call `free', so protect it too. */
  if (!hx_trace_suspend && hx_trace_file)
    {
    fprintf(hx_trace_file, "- %#x %#x\n", caller, ptr);
    hx_draw_stack_memory();
    }

  HX_UNLOCK;
}
